import json
import sys
from wsgiref import simple_server

import falcon
from oslo_config import cfg
from oslo_config.cfg import StrOpt, IntOpt
from oslo_log import log

from orchestrator.model.resources import ResourceGroup, InvalidResourceException, Resource

HTTP_SERVER_OPTS = [
    StrOpt('host', default='127.0.0.1'),
    IntOpt('port', default=8000)
]
cfg.CONF.register_cli_opts(HTTP_SERVER_OPTS, 'http')

log.register_options(cfg.CONF)
log.setup(cfg.CONF, 'orchestrator')


class Stack(object):
    def __init__(self):
        self.logger = log.getLogger()
        self.stacks = {}

    def on_get(self, _request, response, stack_id=None):
        if stack_id is not None:
            if stack_id in self.stacks:
                response.body = json.dumps(self.stacks[stack_id], default=Resource.json_encoder)
            else:
                raise falcon.HTTPNotFound()
        response.body = json.dumps(list(self.stacks.values()), default=Resource.json_encoder)

    def on_post(self, request, response):
        if request.content_length:
            req = request.stream.read().decode('utf8')
            stack = ResourceGroup(json.loads(req))
            flat_stack = stack.flatten(flat_leaves=False)
            self.__deploy_resources(flat_stack)
            self.stacks[stack.id] = stack
            response.body = str(stack)
            response.status = falcon.HTTP_CREATED
        else:
            raise falcon.HTTPBadRequest('Empty request received')

    def on_put(self, request, response, stack_id):
        if stack_id in self.stacks.keys():
            update_body = request.stream.read().decode('utf8')
            updated_stack = json.loads(update_body)
            diff = self.stacks[stack_id].diff(ResourceGroup(updated_stack), flat_leaves=False)
            self.__deploy_resources(diff['added'])
            response.body = json.dumps(diff, default=Resource.json_encoder)
            self.stacks[stack_id] = updated_stack
        else:
            raise falcon.HTTPNotFound()

    def __deploy_resources(self, flat_stack):
        results = {}
        for fqn, resource in flat_stack.items():
            results[fqn] = resource.deploy(fqn)
            self.logger.info('Deployment started: %s', fqn)
        for fqn, result in results.items():
            if result is not None:
                result.wait(timeout=cfg.CONF.rpc.operation_timeout)
                self.logger.info('Deployment finished: %s (%s)', fqn, result.status)
            else:
                self.logger.info('Deployment status N/A of %s', fqn)

    def on_delete(self, _request, response, stack_id):
        if stack_id in self.stacks.keys():
            del self.stacks[stack_id]
            response.status = falcon.HTTP_NO_CONTENT
        else:
            raise falcon.HTTPNotFound()


class LoggingMiddleware(object):
    logger = log.getLogger()

    def process_request(self, request, *_):
        self.logger.info('Request received: %s %s', request.method, request.path)

    def process_response(self, _request, response, *_):
        if response.body:
            self.logger.info('Sending response (%s): %s', response.status, response.body)


def handle_method_not_allowed(*args):
    params = args[3]
    raise falcon.HTTPMethodNotAllowed(['GET', 'PUT', 'DELETE'] if params else ['GET', 'POST'])


def handle_bad_request(*args):
    exception = args[0]
    raise falcon.HTTPBadRequest(description=str(exception))


def create_app():
    stack_controller = Stack()
    app = falcon.API(middleware=[LoggingMiddleware()])
    app.add_route('/v1/stacks', stack_controller)
    app.add_route('/v1/stacks/{stack_id}', stack_controller)
    app.add_error_handler(TypeError, handle_method_not_allowed)
    app.add_error_handler(InvalidResourceException, handle_bad_request)
    return app


def main(args=None):
    cfg.CONF(args[1:] if args else None)
    app = create_app()
    httpd = simple_server.make_server(cfg.CONF.http.host, cfg.CONF.http.port, app)
    httpd.serve_forever()


if __name__ == '__main__':
    main(sys.argv)
