import json
import uuid
from socket import inet_aton

from enum import Enum
from oslo_config import cfg
from voluptuous import Schema, Invalid, MultipleInvalid, Any, ALLOW_EXTRA, PREVENT_EXTRA, re, Required, Optional, All, \
    Range

from orchestrator.util.task_client import Tasks


class Resource(object):
    DEFAULT_ID_LENGTH = 8
    ADDITIONAL_SCHEMA = {}

    def __new__(cls, *args, **kwargs):
        attributes = cls.__attributes_from_dict(args[0] if len(args) else kwargs)
        resource_type = attributes['type'].value
        if resource_type is Resource:
            raise InvalidResourceException('Resource class is abstract')
        cls.__validate_attributes(attributes, cls.__schema(resource_type, extend=True))
        instance = super(Resource, resource_type).__new__(resource_type)
        instance._attributes = attributes  # pylint: disable=protected-access
        return instance

    @classmethod
    def __attributes_from_dict(cls, attributes):
        cls.__validate_attributes(attributes, cls.__schema(extend=False))
        attributes = attributes.copy()
        attributes['id'] = attributes.get('id', Resource.__generate_random_id())
        attributes['type'] = ResourceType.validate(attributes.get('type', cls))
        return attributes

    @classmethod
    def __validate_attributes(cls, attributes, schema):
        try:
            schema(attributes)
        except MultipleInvalid as invalid:
            raise InvalidResourceException(str(invalid))

    @classmethod
    def __schema(cls, resource_type=None, extend=False):
        schema = Schema(schema={'id': cls._id_validator, 'type': ResourceType.validate}, extra=ALLOW_EXTRA)
        if extend:
            additional = resource_type.ADDITIONAL_SCHEMA if resource_type else cls.ADDITIONAL_SCHEMA
            schema = schema.extend(additional, extra=PREVENT_EXTRA)
        return schema

    @staticmethod
    def _id_validator(resource_id):
        if not re.compile('^[a-zA-Z0-9_-]{2,42}$').match(resource_id):
            raise Invalid("'%s' is invalid" % resource_id)

    @staticmethod
    def _mac_validator(mac):
        return re.match("[0-9a-f]{2}([-:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$", mac.lower())

    @staticmethod
    def _ip_validator(ip_address):
        try:
            return inet_aton(ip_address)
        except AttributeError:
            return False

    @property
    def type(self):
        return str(self._attributes['type'])

    def __getattr__(self, item):
        try:
            return self._attributes[item]
        except KeyError:
            raise AttributeError('Attribute (%s) is not set on resource')

    @staticmethod
    def __generate_random_id():
        return uuid.uuid4().hex[:Resource.DEFAULT_ID_LENGTH]

    def __eq__(self, other):
        try:
            return self.flatten(flat_leaves=True) == other.flatten(flat_leaves=True)
        except AttributeError:
            return False

    def __repr__(self):
        return json.dumps(self, default=Resource.json_encoder)

    def __iter__(self):
        return self._attributes.__iter__()

    def keys(self):
        return self._attributes.keys()

    def values(self):
        return self._attributes.values()

    def items(self):
        return self._attributes.items()

    def __getitem__(self, item):
        return self._attributes[item]

    def __add__(self, other):
        return ResourceGroup(resources=[self, other])

    def flatten(self, fqn=None, flat_leaves=True):
        fqn = '%s.%s' % (fqn, self.id) if fqn else self.id
        if flat_leaves:
            leaves = {}
            for key, value in self.items():
                leaves['%s.%s' % (fqn, key)] = str(value) if isinstance(value, ResourceType) else value
            return leaves
        return {fqn: self}

    @staticmethod
    def json_encoder(data):
        if isinstance(data, ResourceType):
            return str(data)
        return dict(data)

    def deploy(self, name):
        pass


class ResourceGroup(Resource):
    ADDITIONAL_SCHEMA = {
        Optional('resources'): Any(list, set, tuple)
    }

    def __init__(self, *_, **__):
        self.__resources = {}
        if 'resources' in self._attributes.keys():
            for resource in self._attributes['resources']:
                self.add(resource)
        self._attributes['resources'] = self.resources

    def __eq__(self, other):
        try:
            return self.id == other.id and super(ResourceGroup, self).__eq__(other)
        except AttributeError:
            return False

    def add(self, resource):
        if not isinstance(resource, Resource):
            try:
                resource = Resource(resource)
            except InvalidResourceException:
                raise InvalidResourceException('ResourceGroup can only store resources')
        if resource.id not in self.__resources.keys():
            self.__resources[resource.id] = resource
        else:
            raise KeyError("ResourceGroup '%s' already contains a resource with id '%s'" % (self.id, resource.id))

    def remove(self, resource):
        try:
            del self.__resources[Resource(resource).id]
        except InvalidResourceException:
            del self.__resources[resource.id]

    @property
    def resources(self):
        return list(self.__resources.values())

    def diff(self, other, flat_leaves=True):
        diff = {'added': {}, 'removed': {}}
        flat_self = self.flatten(flat_leaves=flat_leaves)
        flat_other = other.flatten(flat_leaves=flat_leaves)
        diff_keys = set(flat_self.keys()).symmetric_difference(flat_other.keys())
        for key in diff_keys:
            if key in flat_self.keys():
                diff['removed'][key] = flat_self[key]
            if key in flat_other.keys():
                diff['added'][key] = flat_other[key]
        return diff

    def flatten(self, fqn=None, flat_leaves=True):
        flat = dict()
        for resource in self.resources:
            flat.update(resource.flatten('%s.%s' % (fqn, self.id) if fqn else self.id, flat_leaves))
        return flat


class Instance(Resource):
    ADDITIONAL_SCHEMA = {
        Required('cpu_count'): int,
        Required('memory'): int
    }

    def deploy(self, name):
        args = {
            'name': name,
            'vcpu': self.cpu_count,
            'memory_max': self.memory
        }
        result = Tasks.send_task('vmdriver.create', args=[args])
        return result


class Disk(Resource):
    ADDITIONAL_SCHEMA = {
        Required('size'): All(int, Range(min=1)),
        Optional('base'): str
    }

    def deploy(self, name):
        args = {
            'dir': cfg.CONF.storage.datastore_root,
            'name': name,
            'format': 'qcow2',
            'type': 'normal',
            'size': self.size,
            'base_name': self.base or None
        }
        task_name = 'storagedriver.snapshot' if self.base else 'storagedriver.create'
        result = Tasks.send_task(task_name, args=[args])
        return result


class DiskAttachment(Resource):
    ADDITIONAL_SCHEMA = {
        Required('instance_id'): Resource._id_validator,
        Required('stack_id'): Resource._id_validator
    }


class NetworkConnection(Resource):
    ADDITIONAL_SCHEMA = {
        Required('instance_id'): Resource._id_validator,
        Required('mac'): Resource._mac_validator,
        Required('vlan'): All(int, Range(min=1, max=4094)),
        Required('ip_address'): Resource._ip_validator
    }


class ResourceType(Enum):
    null = Resource
    group = ResourceGroup
    instance = Instance
    disk = Disk
    attachment = DiskAttachment
    connection = NetworkConnection

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        try:
            return cls[value]
        except KeyError:
            return cls(value)
        except ValueError:
            raise Invalid("'%s' is invalid" % value)

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, str(self))


class InvalidResourceException(Exception):
    pass
