import copy
import json
from unittest.case import TestCase

from orchestrator.model.resources import Resource, ResourceGroup, Instance, InvalidResourceException, ResourceType

TEST_INSTANCE_ARGS = {
    'cpu_count': 2,
    'memory': 256 * 1000 * 1000
}


class ResourceTest(TestCase):
    def test_resource_without_args(self):
        with self.assertRaises(InvalidResourceException):
            Resource()

    def test_resource_from_args(self):
        # given
        expected_id = 'dummy_resource'
        expected_type = ResourceType.instance

        # when
        result = Resource(id=expected_id, type=expected_type, **TEST_INSTANCE_ARGS)

        # then
        self.assertEqual(expected_id, result.id)
        self.assertEqual(expected_type.name, result.type)

    def test_resource_from_dict(self):
        # given
        resource_dict = {'id': 'dummy_resource', 'type': 'instance'}
        resource_dict.update(TEST_INSTANCE_ARGS)

        # when
        result = Resource(resource_dict)

        # then
        self.assertEqual(resource_dict['id'], result.id)
        self.assertEqual(resource_dict['type'], result.type)

    def test_resource_from_invalid_args(self):
        with self.assertRaises(InvalidResourceException):
            Resource(id='dummy_id', invalid_attribute=False)

    def test_resource_from_invalid_dict(self):
        # given
        invalid_attributes = dict(id='dummy_id', invalid_attribute=False)

        # when/then
        with self.assertRaises(InvalidResourceException):
            Resource(invalid_attributes)

    def test_resource_id_validation(self):
        invalid_ids = [
            'resource_id_which_is_way_too_long_to_be_valid',
            'some.resource.id',
            'some resource id',
            'a'
        ]

        for invalid_id in invalid_ids:
            with self.assertRaises(InvalidResourceException):
                Instance(id=invalid_id, **TEST_INSTANCE_ARGS)

        self.assertIsInstance(Instance(id='valid_id', **TEST_INSTANCE_ARGS), Instance)
        self.assertIsInstance(Instance(id='valid-id', **TEST_INSTANCE_ARGS), Instance)
        self.assertIsInstance(Instance(id='VALID', **TEST_INSTANCE_ARGS), Instance)
        self.assertIsInstance(Instance(id='123456', **TEST_INSTANCE_ARGS), Instance)
        self.assertIsInstance(Instance(id='_id-42', **TEST_INSTANCE_ARGS), Instance)

    def test_resource_type_validation(self):
        self.assertTrue(ResourceType.validate('instance'))
        self.assertTrue(ResourceType.validate(Instance))
        self.assertTrue(ResourceType.validate(ResourceType.instance))
        self.assertIsInstance(Resource(type='instance', **TEST_INSTANCE_ARGS), Instance)
        self.assertIsInstance(Resource(type=Instance, **TEST_INSTANCE_ARGS), Instance)
        self.assertIsInstance(Resource(type=ResourceType.instance, **TEST_INSTANCE_ARGS), Instance)
        self.assertEqual(Instance(id='some_instance', **TEST_INSTANCE_ARGS),
                         Resource(id='some_instance', type='instance', **TEST_INSTANCE_ARGS))
        self.assertEqual(Instance(id='some_instance', **TEST_INSTANCE_ARGS),
                         Resource(id='some_instance', type=Instance, **TEST_INSTANCE_ARGS))
        self.assertEqual(Instance(id='some_instance', **TEST_INSTANCE_ARGS),
                         Resource(id='some_instance', type=ResourceType.instance, **TEST_INSTANCE_ARGS))

    def test_resource_type_invalid(self):
        with self.assertRaises(ValueError):
            ResourceType.validate('nonexisting_resource_type')

    def test_resource_instance(self):
        # when
        instance = Instance(TEST_INSTANCE_ARGS)

        # then
        self.assertIsInstance(instance.id, str)
        self.assertEqual(Resource.DEFAULT_ID_LENGTH, len(instance.id))
        self.assertEqual(ResourceType.instance.name, instance.type)

    def test_resource_equivalence(self):
        # given
        resource1 = Resource(id='some_resource', type=Instance, **TEST_INSTANCE_ARGS)
        resource2 = Resource(id='some_resource', type=Instance, **TEST_INSTANCE_ARGS)
        resource3 = Instance(id='some_resource', **TEST_INSTANCE_ARGS)
        resource4 = Resource(id='some_other_resource', type=Instance, **TEST_INSTANCE_ARGS)
        resource5 = Resource(id='some_resource', type=ResourceGroup)

        # then
        self.assertEqual(resource1, resource1)
        self.assertEqual(resource1, resource2)
        self.assertEqual(resource1, resource3)
        self.assertNotEqual(resource1, resource4)
        self.assertNotEqual(resource1, resource5)
        self.assertFalse(resource1 == 500)

    def test_resource_str_representation(self):
        # given
        expected_json = '{"id": "some_resource", "type": "instance"}'
        expected_dict = json.loads(expected_json)
        expected_dict.update(TEST_INSTANCE_ARGS)

        # when
        resource = Resource(expected_dict)
        resource_dict = json.loads(str(resource))
        resource_dict.update(TEST_INSTANCE_ARGS)

        # then
        self.assertEqual(expected_dict, resource_dict)

    def test_resource_dict_representation(self):
        # given
        expected_dict = {'id': 'some_resource', 'type': ResourceType.instance}
        expected_dict.update(TEST_INSTANCE_ARGS)

        # when
        resource = Resource(expected_dict)
        resource_dict = dict(resource)

        # then
        self.assertEqual(expected_dict, resource_dict)

    def test_resource_dynamic_type(self):
        # when
        args = dict(type='instance', **TEST_INSTANCE_ARGS)

        # then
        self.__assert_is_instance(Resource(args))
        self.__assert_is_instance(Resource(**args))
        self.__assert_is_instance(Instance(**args))

        with self.assertRaises(InvalidResourceException):
            Resource()

    def __assert_is_instance(self, resource):
        self.assertIsInstance(resource, Instance)
        self.assertEqual(ResourceType.instance.name, resource.type)


class ResourceGroupTest(TestCase):
    def test_resource_group_without_args(self):
        # when
        empty_group = ResourceGroup()

        # then
        self.assertEqual([], empty_group.resources)

    def test_resource_group_from_resources(self):
        # given
        resource1 = Instance(id='some_resource', **TEST_INSTANCE_ARGS)
        resource2 = Instance(id='some_other_resource', **TEST_INSTANCE_ARGS)

        # then
        self.assertEqual([resource1], ResourceGroup(resources=[resource1]).resources)
        for resources in [[resource1, resource2], [resource2, resource1]]:
            self.assertEqual(2, len(ResourceGroup(resources=resources).resources))
            self.assertIn(resource1, ResourceGroup(resources=resources).resources)
            self.assertIn(resource2, ResourceGroup(resources=resources).resources)

    def test_resource_group_from_invalid_collection(self):
        with self.assertRaises(InvalidResourceException):
            ResourceGroup(resources={'invalid': 'resource'})

    def test_resource_group_from_dicts(self):
        # given
        dict1 = dict(id='some_resource', type=Instance, **TEST_INSTANCE_ARGS)
        dict2 = dict(id='some_other_resource', type=Instance, **TEST_INSTANCE_ARGS)
        expected1 = Instance(dict1)
        expected2 = Instance(dict2)

        # then
        self.assertEqual([expected1], ResourceGroup(resources=[dict1]).resources)
        for resources in [[expected1, expected2], [dict2, dict1]]:
            self.assertEqual(2, len(ResourceGroup(resources=resources).resources))
            self.assertIn(expected1, ResourceGroup(resources=resources).resources)
            self.assertIn(expected2, ResourceGroup(resources=resources).resources)

    def test_resource_group_from_dict(self):
        # given
        expected_group_id = 'some_resource_group'
        expected_resource_id = 'some_instance'
        group_dict = {
            'id': expected_group_id,
            'type': 'group',
            'resources': [
                dict(id=expected_resource_id, type='instance', **TEST_INSTANCE_ARGS)
            ]
        }

        # when
        group = ResourceGroup(group_dict)

        # then
        self.assertEqual(expected_group_id, group.id)
        self.assertEqual(1, len(group.resources))
        self.assertEqual(expected_resource_id, group.resources[0].id)

    def test_resource_group_add_resource(self):
        # given
        group = ResourceGroup()
        resource = Instance(TEST_INSTANCE_ARGS)

        # when
        group.add(resource)

        # then
        self.assertEqual(1, len(group.resources))
        self.assertIn(resource, group.resources)

    def test_resource_group_remove_resource(self):
        # given
        resource = Instance(TEST_INSTANCE_ARGS)
        group = ResourceGroup()
        group.add(resource)

        # when
        group.remove(resource)

        # then
        self.assertEqual([], group.resources)

    def test_resource_group_duplicate_resource_id(self):
        # given
        group = ResourceGroup()
        resource = Instance(TEST_INSTANCE_ARGS)

        # then
        with self.assertRaises(KeyError):
            group.add(resource)
            group.add(resource)

    def test_resource_group_addition(self):
        # given
        group1 = ResourceGroup()
        group2 = ResourceGroup()
        resource = Instance(TEST_INSTANCE_ARGS)

        # when
        result_group1 = group1 + group2
        result_group2 = result_group1 + resource

        # then
        self.assertIsInstance(result_group1, ResourceGroup)
        self.assertIsInstance(result_group2, ResourceGroup)
        self.assertEqual(2, len(result_group1.resources))
        self.assertEqual(2, len(result_group2.resources))
        self.assertIn(group1, result_group1.resources)
        self.assertIn(group2, result_group1.resources)
        self.assertIn(resource, result_group2.resources)

    def test_resource_group_iteration(self):
        # when
        group = ResourceGroup(resources=[Instance(TEST_INSTANCE_ARGS),
                                         Instance(TEST_INSTANCE_ARGS),
                                         Instance(TEST_INSTANCE_ARGS)])

        # then
        for index, resource in enumerate(group.resources):
            self.assertEqual(group.resources[index], resource)

    def test_resource_group_equivalence(self):
        # given
        resource1 = Instance(id='some_resource', **TEST_INSTANCE_ARGS)
        resource2 = Instance(id='some_other_resource', **TEST_INSTANCE_ARGS)

        # then
        self.assertEqual(ResourceGroup(id='some_group'),
                         ResourceGroup(id='some_group'))
        self.assertEqual(ResourceGroup(id='some_group', resources=[resource1]),
                         ResourceGroup(id='some_group', resources=[resource1]))
        self.assertNotEqual(ResourceGroup(id='some_group'),
                            ResourceGroup(id='some_other_group'))
        self.assertNotEqual(ResourceGroup(id='some_group', resources=[resource1]),
                            ResourceGroup(id='some_group', resources=[resource2]))

    TEST_TREE_DICT = {
        'id': 'root',
        'type': 'group',
        'resources': [
            {
                'id': 'instance-1',
                'type': 'instance',
                'cpu_count': 1,
                'memory': 100
            },
            {
                'id': 'level-1',
                'type': 'group',
                'resources': [
                    {
                        'id': 'instance-2',
                        'type': 'instance',
                        'cpu_count': 2,
                        'memory': 200
                    },
                    {
                        'id': 'instance-3',
                        'type': 'instance',
                        'cpu_count': 3,
                        'memory': 300
                    }
                ]
            }
        ]
    }

    def test_resource_group_flatten_flat_leaves(self):
        # given
        expected_flat_dict = {
            'root.instance-1.id': 'instance-1',
            'root.instance-1.type': 'instance',
            'root.instance-1.cpu_count': 1,
            'root.instance-1.memory': 100,
            'root.level-1.instance-2.id': 'instance-2',
            'root.level-1.instance-2.type': 'instance',
            'root.level-1.instance-2.cpu_count': 2,
            'root.level-1.instance-2.memory': 200,
            'root.level-1.instance-3.id': 'instance-3',
            'root.level-1.instance-3.type': 'instance',
            'root.level-1.instance-3.cpu_count': 3,
            'root.level-1.instance-3.memory': 300
        }

        # when
        tree_resource = Resource(self.TEST_TREE_DICT)
        flat_dict = tree_resource.flatten()

        # then
        self.assertEqual(expected_flat_dict, flat_dict)

    def test_resource_group_flatten_not_flat_leaves(self):
        # given
        expected_flat_dict = {
            'root.instance-1': Instance(id='instance-1', cpu_count=1, memory=100),
            'root.level-1.instance-2': Instance(id='instance-2', cpu_count=2, memory=200),
            'root.level-1.instance-3': Instance(id='instance-3', cpu_count=3, memory=300)
        }

        # when
        tree_resource = Resource(self.TEST_TREE_DICT)
        flat_dict = tree_resource.flatten(flat_leaves=False)

        # then
        self.assertEqual(expected_flat_dict, flat_dict)

    def test_resource_group_diff_flat_leaves(self):
        # given
        group1 = ResourceGroup(self.TEST_TREE_DICT)

        tree_dict2 = copy.deepcopy(self.TEST_TREE_DICT)
        del tree_dict2['resources'][0]
        group2 = ResourceGroup(tree_dict2)

        expected_diff_values = {
            'root.instance-1.id': 'instance-1',
            'root.instance-1.type': 'instance',
            'root.instance-1.cpu_count': 1,
            'root.instance-1.memory': 100
        }
        expected_diff_added = {
            'added': expected_diff_values,
            'removed': {}
        }
        expected_diff_removed = {
            'added': {},
            'removed': expected_diff_values
        }

        # when
        diff1 = group1.diff(group2, flat_leaves=True)
        diff2 = group2.diff(group1, flat_leaves=True)

        # then
        self.assertNotEqual(diff1, diff2)
        self.assertEqual(expected_diff_removed, diff1)
        self.assertEqual(expected_diff_added, diff2)

    def test_resource_group_diff_not_flat_leaves(self):
        # given
        group1 = ResourceGroup(self.TEST_TREE_DICT)

        tree_dict2 = copy.deepcopy(self.TEST_TREE_DICT)
        removed_resource = tree_dict2['resources'][0]
        del tree_dict2['resources'][0]
        group2 = ResourceGroup(tree_dict2)

        expected_diff_values = {
            'root.instance-1': Instance(id=removed_resource['id'], cpu_count=1, memory=100)
        }
        expected_diff_added = {
            'added': expected_diff_values,
            'removed': {}
        }
        expected_diff_removed = {
            'added': {},
            'removed': expected_diff_values
        }

        # when
        diff1 = group1.diff(group2, flat_leaves=False)
        diff2 = group2.diff(group1, flat_leaves=False)

        # then
        self.assertNotEqual(diff1, diff2)
        self.assertEqual(expected_diff_removed, diff1)
        self.assertEqual(expected_diff_added, diff2)
